#############################################################################
#
# Author: Michel F. SANNER
# Reimplemented from Babel v1.6 from Pat Walters and Math Stahl
#
# Copyright: M. Sanner TSRI 2000
#
#############################################################################
#
# $Header: /opt/cvs/python/packages/share1.5/PyBabel/__init__.py,v 1.10 2006/03/15 00:45:52 annao Exp $
#
# $Id: __init__.py,v 1.10 2006/03/15 00:45:52 annao Exp $
#
#

"""

PyBabel v0.1alpha

    This a Python re-implmentation of some of Babel v1.6 functionalities, including:

    - Atom hybridization assignment (from geometry)
    - Gasteiger charges calculations
    - atom type conversions
    - properties from atom element name
    - finding rings
    - assigning bond orders
    - identifying aromatic rings
    - adding hydrogen atoms




I - Atom hybridization assignment:

    The AtomHybridization class allows to assign atom types.
  
  example:
    
    >>> #create an instance of AtomHybridization
    >>> from PyBabel.atomTypes import AtomHybridization
    >>> babel = AtomHybridization()
    >>> babel.assignHybridization(allAtoms)

    allAtoms is a list of 'Atom' objects having the following members:

        - a.element : atom's chemical element symbol (string,
                      2 character max, second character lower case)

        - a.coords  : 3-sequence of floats (tuple, float or array)
  
        - a.bonds   : list of Bond objects

    Bond are objects having the following members:

        - b.atom1 : instance of Atom

        - b.atom2 : instance of Atom


  After completion each atom has the following additional members:

      - babel_type          : string
      - babel_atomic_number : int
      - babel_organic       : 0/1




II - Gasteiger charges calculations:

  Before Gasteiger charges can be calculated, babel AtomHybridization as to
  be assigned (see I).

  example:
    
      >>> #create an instance of AtomHybridization
      >>> from PyBabel.atomTypes import AtomHybridization
      >>> babel = AtomHybridization()
      >>> babel.assignHybridization(allAtoms)
      >>>
      >>> from PyBabel.gasteiger import Gasteiger
      >>> Gast = Gasteiger()
      >>> Gast.compute(allAtoms)


  allAtoms is a list of 'Atom' objects having the following members:

      * Atom:

          - a.babel_type

          - a.babel_atomic_number

          - a.bonds

  After completion each atom has a gast_charge member




III - types conversion

  Before atom types can be obtained, babel's internal types have to be assigned
  (see I)

  example:
    
      >>> #create an instance of AtomHybridization
      >>> from PyBabel.atomTypes import AtomHybridization
      >>> babel = AtomHybridization()
      >>> babel.assignHybridization(allAtoms)
      >>>
      >>> from PyBabel.atomTypes import AtomHybridization
      >>> converter = TypeConverter('SYB')
      >>> newTypes = map( converter.convert, allAtoms.babel_type)




IV - properties from atom element name

  Before atom properties can be obtained, babel's internal types have to 
  be assigned (see I)

  example:
    
      >>> #create an instance of AtomHybridization
      >>> from PyBabel.atomTypes import AtomHybridization
      >>> babel = AtomHybridization()
      >>> babel.assignHybridization(allAtoms)
      >>>
      >>> rad = babel.getProperty('vdw_rad', allAtoms.element)

  valid property names are:

      - 'num'

      - 'cov_rad'

      - 'bond_ord_rad'

      - 'vdw_rad'

      - 'bs_rad'

      - 'max_bonds'

      - 'rgb'




V - finding rings:

  This algorithm is different from the one used in Babel to avoid the N^2
  complexity in number of bonds.

  When rings are nested the smalest rings are reported.

  The algorithms might fail to report all rings if all the atoms in some
  rings belong to multiple rings.

  example:
    
      >>> #create an instance of RingFinder object
      >>> from PyBabel.cycle import RingFinder
      >>> r = RingFinder()
      >>> r.findRings(atoms, bonds)
      >>> r.printRings()

  atoms has to be a list of Atom objects

      * Atom:
          
          - a.coords : 3-sequence of floats
          
          - a.bonds : list of Bond objects
          
      * Bond:
          
          - b.atom1 : instance of Atom
          
          - b.atom2 : instance of Atom
          

  After completion the RingFinder object has the following members:
      
      - ringCount: number of rings
      
      - rings : a list of rings. A ring is a dictionary with 2 keys
              'atoms' and 'bonds'
              
      - allRingAtoms: list of atoms that are in rings(atoms may appear twice)

      - allRingBonds: list of bonds that are in rings(bonds may appear twice)


  In addition:

      atoms involved in rings have a member 'ring' that is a list of
      rings they belong to (0-based list of integers)




VI - assigning bond orders:

  The BondOrder class that can be used to compute bond order.

  Before a BondOrder object can be used, atoms must have been assigned
  a type (see I)

  Bond order can be calculated using 2 different methods depending on whether
  rings have been identified previously or not. Babel decides to use the first
  method for molecules with more than 200 atoms and the second one else.

  example:
    
      >>> #create an instance of AtomHybridization
      >>> from PyBabel.atomTypes import AtomHybridization
      >>> atype = AtomHybridization()
      >>> atype.assignHybridization(atoms)

      >>> #create an instance of BondOrder object
      >>> from PyBabel.bo import BondOrder
      >>> bo = BondOrder()
      >>> bo.assignBondOrder( atoms, bonds )

      or

      >>> #create an instance of AtomHybridization
      >>> from PyBabel.atomTypes import AtomHybridization
      >>> atype = AtomHybridization()
      >>> atype.assignHybridization(atoms)

      >>> from PyBabel.cycle import RingFinder
      >>> rings = RingFinder()
      >>> rings.findRings(allAtoms, bonds)

      >>> from PyBabel.bo import BondOrder
      >>> bo = BondOrder()
      >>> bo.assignBondOrder( atoms, bonds, rings )

  
  atoms has to be a list of atom objects

      * Atom:
          
          - a.coords : 3-sequence of floats
          
          - a.bonds : list of Bond objects
          
          - babel_type: string
          
          - babel_atomic_number: int

      * Bond:
          
          - b.atom1 : instance of Atom
          
          - b.atom2 : instance of Atom
          

  After completion each bond has a 'bondOrder' attribute (integer)




VII - identifying aromatic rings

  Before this Aromatic object can be used to identify aromatic rings,
  
        1. atoms must have been assigned a type (see I)

	2. rings must have been identified (see V)

	3. bond orders must have been defined (see VI)

  example:
    
      >>> #create an instance of AtomHybridization
      >>> from PyBabel.atomTypes import AtomHybridization
      >>> atype = AtomHybridization()
      >>> atype.assignHybridization(atoms)

      >>> from PyBabel.cycle import RingFinder
      >>> rings = RingFinder()
      >>> rings.findRings(atoms, bonds)

      >>> from PyBabel.bo import BondOrder
      >>> bo = BondOrder()
      >>> bo.assignBondOrder(atoms, bonds, rings)

      >>> from PyBabel.aromatic import Aromatic
      >>> arom = Aromatic(rings)
      >>> arom.find_aromatic_atoms(atoms)

  atoms has to be a list of atom objects

      * Atom:

          - a.coords : 3-sequence of floats

          - a.bonds : list of Bond objects

          - babel_type: string

          - babel_atomic_number: int

      * Bond:

          - b.atom1 : instance of Atom

          - b.atom2 : instance of Atom

  After completion the bondOrder member of bonds in aromatic bonds is
  is 'aromatic'. 

  The aromatic rings has a new key called 'aromatic' set to 1 or 0
	



VIII - adding hydrogen atoms

  Before this AddHydrogens object can be used, atoms must have been assigned
  a type (see I)

  Hydrogen atoms can be added using 2 different methods. The first one 
  requires bondOrders to have been calculated previousely.

  example:
    
      >>> #create an instance of AtomHybridization
      >>> from PyBabel.atomTypes import AtomHybridization
      >>> atype = AtomHybridization()
      >>> atype.assignHybridization(atoms)

      >>> from PyBabel.addh import AddHydrogens
      >>> addh = AddHydrogens()
      >>> hat = addh.addHydrogens(atoms)

  atoms has to be a list of atom objects:

      * Atom:
          
          - a.coords : 3-sequence of floats

          - a.bonds : list of Bond objects

          - babel_type: string

          - babel_atomic_number: int

      * Bond:
          
          - b.atom1 : instance of Atom

          - b.atom2 : instance of Atom

  example:
  
      >>> from PyBabel.addh import AddHydrogens
      >>> addh = AddHydrogens()
      >>> hat = addh.addHydrogens(atoms, method='withBondOrder')

  atoms has to be a list of atom objects as above and

       * Bond:
           
           - b.atom1 : instance of Atom

           - b.atom2 : instance of Atom

           - b.bondOrder : integer

          
  These calls return a list 'hat' containing a tuple for each Hydrogen
  atom to be added. This tuple provides:

      - coordsH       : 3-float coordinates of the H atom
      
      - atom          : the atom to which the H atom is to be connected
      
      - atomic_number : the babel_atomic_number of the H atom
      
      - type          : tyhe babel_type of the H atom





Michel Sanner April 2000
sanner@scripps.edu
"""

CRITICAL_DEPENDENCIES = ['MolKit', 'mglutil']
NONCRITICAL_DEPENDENCIES = ['Pmv', 'DejaVu']



#############################################################################
#
# Author: Michel F. SANNER
# Reimplemented from Babel v1.6 from Pat Walters and Math Stahl
#
# Copyright: M. Sanner TSRI 2000
#
#############################################################################
#
# $Header: /opt/cvs/python/packages/share1.5/PyBabel/addh.py,v 1.6 2007/10/11 17:43:48 sargis Exp $
#
# $Id: addh.py,v 1.6 2007/10/11 17:43:48 sargis Exp $
#

"""
This file implements the AddHydrogens class.

Before this AddHydrogens object can be used, atoms must have been assigned
a type see (AtomHybridization in types.py).

Hydrogen atoms can be added using 2 different methods. The first one requires
bondOrders to have been calculated previousely.

example:
    
      >>> atype = AtomHybridization()
      >>> atype.assignHybridization(atoms)
      >>> addh = AddHydrogens()
      >>> hat = addh.addHydrogens(atoms)

      atoms has to be a list of atom objects
      Atom:
          a.coords : 3-sequence of floats
          a.bonds : list of Bond objects
          babel_type: string
          babel_atomic_number: int

      Bond:
          b.atom1 : instance of Atom
          b.atom2 : instance of Atom

      or
      
      >>> addh = AddHydrogens()
      >>> hat = addh.addHydrogens(atoms, method='withBondOrder')

      atoms has to be a list of atom objects as above and
      Bond:
          b.atom1 : instance of Atom
          b.atom2 : instance of Atom
          b.bondOrder : integer
          
      these calls return a list 'hat' containing a tuple for each Hydrogen
      atom to be added. This tuple provides:
          coordsH       : 3-float coordinates of the H atom
          atom          : the atom to which the H atom is to be connected
          atomic_number : the babel_atomic_number of the H atom
          type          : tyhe babel_type of the H atom

reimplmentation of Babel1.6 in Python by Michel Sanner April 2000
Original code by W. Patrick Walters and Matthew T. Stahl 
"""


import math

from atomTypes import TypeConverter
from util import *

ONE_OVER_SQRT3 = 0.577350269
SQRT_TWO_THIRDS = 0.816496581

SP3_C_H_DIST = 1.115
SP2_C_H_DIST = 1.103
SP_C_H_DIST = 1.090

SP3_N_H_DIST = 1.020
SP2_N_H_DIST = 1.020

SP3_O_H_DIST = 0.950

class AddHydrogens:
    """ """

    def addHydrogens(self, atoms, method='noBondOrder'):
        """ """
	Hatoms = []
#        method = 'noBondOrder'
        if method=='noBondOrder':
            num_H_to_add = self.count_missing_hydrogens(atoms)
            if num_H_to_add:
                Hatoms = self.place_hydrogens1(atoms, num_H_to_add)

        else:
            num_H_to_add = self.count_missing_bo_hydrogens(atoms)
            if num_H_to_add:
                Hatoms = self.place_hydrogens2(atoms, num_H_to_add)

            # cleanup
            for a in atoms:
                delattr(a, '_redo')

        return Hatoms


    def place_hydrogens1(self, atoms, num_H_to_add):
        """ """

        Hat = []
        for a in atoms:

            if a.babel_type == "C3":
                val = len(a.bonds)
                if val == 3:
                    Hat = Hat + self.add_tertiary_hydrogen(a, SP3_C_H_DIST)
                elif val == 2:
                    Hat = Hat + self.add_methylene_hydrogens(a ,SP3_C_H_DIST)
                elif val==1:
                    Hat = Hat + self.add_methyl_hydrogen(a, SP3_C_H_DIST)
                    Hat = Hat + self.add_methylene_hydrogens(a, SP3_C_H_DIST,
                                                             Hat[-1])
                    
            elif a.babel_type == "N3+": 
                val = len(a.bonds)
                if val == 2:
                    Hat = Hat + self.add_methylene_hydrogens(a ,SP3_N_H_DIST)
                elif val==1:
                    Hat = Hat + self.add_methyl_hydrogen(a, SP3_N_H_DIST)
                    Hat = Hat + self.add_methylene_hydrogens(a ,SP3_N_H_DIST,
                                                             Hat[-1])

            elif a.babel_type == "C2" or a.babel_type == "Car":
                val = len(a.bonds)
                if val == 2:
                    Hat = Hat + self.add_sp2_hydrogen(a ,SP2_C_H_DIST)
                
            elif a.babel_type == "Npl" or a.babel_type == "Nam" or \
                 a.babel_type == "Ng+":
                val = len(a.bonds)
                if val == 2:
                    Hat = Hat + self.add_sp2_hydrogen(a ,SP2_N_H_DIST)

            elif a.babel_type == "C1":
                if len(a.bonds) == 1:
                    Hat = Hat + self.add_sp_hydrogen(a ,SP_C_H_DIST)

            elif a.babel_type == "O3":
                if len(a.bonds) == 1:
                    Hat = Hat + self.add_methyl_hydrogen(a, SP3_O_H_DIST)


        for a in atoms:

            if a.babel_type == "C2":
                if len(a.bonds) == 1:
                    Hat = Hat + self.add_vinyl_hydrogens(a ,SP2_C_H_DIST)

            elif a.babel_type == "Npl" or a.babel_type == "Nam" or \
                 a.babel_type == "Ng+":
                if len(a.bonds) == 1:
                    # FIXME babel C code says SP2_C_H_DIST here ???
                    Hat = Hat + self.add_vinyl_hydrogens(a ,SP2_N_H_DIST)

        return Hat


    def place_hydrogens2(self, atoms, num_H_to_add):
        """ """

        Hat = []
        converter = TypeConverter("HYB")
        
        for a in atoms:
            type_name = converter.convert(a.babel_type, 'zero')
            hyb = int(type_name)
            code = a.babel_atomic_number * 10 + hyb
            to_add = a._redo
            
            if code == 63: # sp3 carbon
                if to_add==1:
                    Hat = Hat + self.add_tertiary_hydrogen(a, SP3_C_H_DIST)
                elif to_add==2:
                    Hat = Hat + self.add_methylene_hydrogens(a ,SP3_C_H_DIST)
                elif to_add==3:
                    Hat = Hat + self.add_methyl_hydrogen(a, SP3_C_H_DIST)
                    Hat = Hat + self.add_methylene_hydrogens(a, SP3_C_H_DIST,
                                                             Hat[-1])
            elif code == 73: # sp3 nitrogen
                if to_add==1:
                    if a.babel_type=="N3+":
                        Hat = Hat + self.add_tertiary_hydrogen(a, SP3_N_H_DIST)
                    else:
                        Hat = Hat + self.add_sp3_N_hydrogen(a, SP3_N_H_DIST)
                elif to_add==2:
                    Hat = Hat + self.add_methylene_hydrogens(a ,SP3_N_H_DIST)
                elif to_add==3:
                    Hat = Hat + self.add_methyl_hydrogen(a, SP3_N_H_DIST)
                    Hat = Hat + self.add_methylene_hydrogens(a ,SP3_N_H_DIST,
                                                             Hat[-1])
                    
            elif code == 62: # sp2 carbon
                if to_add==1:
                    Hat = Hat + self.add_sp2_hydrogen(a ,SP2_C_H_DIST)

            elif code == 72: # sp2 nitrogen
                if to_add==1:
                    Hat = Hat + self.add_sp2_hydrogen(a ,SP2_N_H_DIST)

            elif code == 61: # sp carbon
                if to_add==1:
                    Hat = Hat + self.add_sp_hydrogen(a ,SP_C_H_DIST)

            elif code == 83: # sp3 oxygen
                if to_add==1:
                    Hat = Hat + self.add_methyl_hydrogen(a, SP3_O_H_DIST)

            # save vinyl and amide protons for last,
            # this way we know where to put them

        for a in atoms:

            type_name = converter.convert(a.babel_type, 'zero')
            hyb = int(type_name)
            code = a.babel_atomic_number * 10 + hyb
            to_add = a._redo

            if code == 62: # sp2 carbon
                if to_add==2:
                    Hat = Hat + self.add_vinyl_hydrogens(a ,SP2_C_H_DIST)

            elif code == 72: # sp2 nitrogens
                if to_add==2:
                    Hat = Hat + self.add_vinyl_hydrogens(a ,SP2_N_H_DIST)

        return Hat


    def type_added_hydrogen(self, atom):
        """ return babel_atomic_number and babel_type for adde H atom"""
  
        atomic_number = 1
        if atom.babel_atomic_number == 6:
            htype = "HC"
        else:
            htype = "H"
            
        return atomic_number, htype


    def add_sp3_N_hydrogen(self, atom, b_length):
        """ """
        c2 = atom.bonds[0].atom1
        if c2==atom: c2 = atom.bonds[0].atom2

        if not addedH:
            c3 = atom.bonds[1].atom1
            if c3==atom: c3 = atom.bonds[1].atom2
            c3 = c3.coords
        else:
            c3 = addedH[0]

        c = atom.coords
        v = vec3(c2.coords, c)
        v1 = vec3(c3, c)
        s = [ v[0]+v1[0], v[1]+v1[1], v[2]+v1[2] ]

        n = [ v[1]*v1[2] - v1[1]*v[2],
              v1[0]*v[2] - v[0]*v1[2],
              v[0]*v1[1] - v1[0]*v[1] ]
       
        s = [ s[0]*ONE_OVER_SQRT3, s[1]*ONE_OVER_SQRT3, s[2]*ONE_OVER_SQRT3 ]
        n = [ n[0]*SQRT_TWO_THIRDS, n[1]*SQRT_TWO_THIRDS,n[2]*SQRT_TWO_THIRDS ]

        h1 = [ s[0]+n[0], s[1]+n[1], s[2]+n[2] ]
        mag= b_length / math.sqrt( h1[0]*h1[0] + h1[1]*h1[1] + h1[2]*h1[2])
        cH1 = [c[0] + (h1[0]*mag), c[1] + (h1[1]*mag), c[2] + (h1[2]*mag)]

        atomic_number, type = self.type_added_hydrogen(atom)

        return [(cH1, atom, atomic_number, type)]

            
    def add_methyl_hydrogen(self, atom, b_length):
        """ """

        b = atom.bonds[0]
        c = atom.coords

        # find atom2 such that (1) atom2 is bonded to atom, and
        #                      (2) atom2 != atom
        atom2 = b.atom1
        c2 = b.atom1.coords
        if b.atom1 == atom:
            atom2 = b.atom2
            c2 = b.atom2.coords

        # find c3 such that (1) atom3 is bonded to atom2, and
        #                   (2) atom3 != atom2, and
        #                   (3) atom3 != atom
        if atom2.bonds[0].atom1 == atom:
            # atom has been located among the bonds of atom2, so
            # move onto the next bond
            c3 = atom2.bonds[1].atom1.coords
            if atom2.bonds[1].atom1==atom2:
                # since atom was part of bonds[0], the other atom of
                # this bond is safe to use (ie. not atom or atom2)
                c3 = atom2.bonds[1].atom2.coords
        else:
            c3 = atom2.bonds[0].atom1.coords
            if atom2.bonds[0].atom1==atom2:
                c3 = atom2.bonds[0].atom2.coords
                # but atom2.bonds[0].atom2 might be atom, so check
                if atom2.bonds[0].atom2==atom:
                    # same logic as above...
                    c3 = atom2.bonds[1].atom1.coords
                    if atom2.bonds[1].atom1==atom2:
                        c3 = atom2.bonds[1].atom2.coords

        # FIX ME: this only works if atom2 has reasonable sp3 geometry!!!
        v = vec3(c3, c2, b_length)
        coordsH = [ c[0]+v[0], c[1]+v[1], c[2]+v[2] ]
        
        atomic_number, type = self.type_added_hydrogen(atom)
        return [(coordsH, atom, atomic_number, type)]


    def add_tertiary_hydrogen(self, atom, b_length):
        """ """
        c2 = atom.bonds[0].atom1
        if c2==atom: c2 = atom.bonds[0].atom2

        c3 = atom.bonds[1].atom1
        if c3==atom: c3 = atom.bonds[1].atom2

        c4 = atom.bonds[2].atom1
        if c4==atom: c4 = atom.bonds[2].atom2

        c = atom.coords
        v1 = vec3(c2.coords, c)
        v2 = vec3(c3.coords, c)
        v3 = vec3(c4.coords, c)

        m = invert_3x3( (v1,v2,v3) )

        s = [ m[0][0] + m[0][1] + m[0][2],
              m[1][0] + m[1][1] + m[1][2],
              m[2][0] + m[2][1] + m[2][2] ]

        mag= b_length / math.sqrt( s[0]*s[0] + s[1]*s[1] + s[2]*s[2])
        cH = [c[0]+ (s[0]*mag), c[1] + (s[1]*mag), c[2] + (s[2]*mag)]

        atomic_number, type = self.type_added_hydrogen(atom)
        return [(cH, atom, atomic_number, type)]


    def add_methylene_hydrogens(self, atom, b_length, addedH=None):
        """ """
        c2 = atom.bonds[0].atom1
        if c2==atom: c2 = atom.bonds[0].atom2

        if not addedH:
            c3 = atom.bonds[1].atom1
            if c3==atom: c3 = atom.bonds[1].atom2
            c3 = c3.coords
        else:
            c3 = addedH[0]
            
        c = atom.coords
        v = vec3(c2.coords, c)
        v1 = vec3(c3, c)
        s = [ v[0]+v1[0], v[1]+v1[1], v[2]+v1[2] ]
        
        n = [ v[1]*v1[2] - v1[1]*v[2],
              v1[0]*v[2] - v[0]*v1[2],
              v[0]*v1[1] - v1[0]*v[1] ]

        s = [ s[0]*ONE_OVER_SQRT3, s[1]*ONE_OVER_SQRT3, s[2]*ONE_OVER_SQRT3 ]
        n = [ n[0]*SQRT_TWO_THIRDS, n[1]*SQRT_TWO_THIRDS,n[2]*SQRT_TWO_THIRDS ]

        h1 = [ s[0]+n[0], s[1]+n[1], s[2]+n[2] ]
        mag= b_length / math.sqrt( h1[0]*h1[0] + h1[1]*h1[1] + h1[2]*h1[2])
        cH1 = [c[0] + (h1[0]*mag), c[1] + (h1[1]*mag), c[2] + (h1[2]*mag)]

        h2 = [ s[0]-n[0], s[1]-n[1], s[2]-n[2] ]
        mag= b_length / math.sqrt( h2[0]*h2[0] + h2[1]*h2[1] + h2[2]*h2[2])
        cH2 = [c[0] + (h2[0]*mag), c[1] + (h2[1]*mag), c[2] + (h2[2]*mag)]

        atomic_number, type = self.type_added_hydrogen(atom)
 
        return [(cH1, atom, atomic_number, type),
                (cH2, atom, atomic_number, type)]


    def add_sp2_hydrogen(self, atom, b_length):
        """ """

        c2 = atom.bonds[0].atom1
        if c2==atom: c2 = atom.bonds[0].atom2

        c3 = atom.bonds[1].atom1
        if c3==atom: c3 = atom.bonds[1].atom2

        c = atom.coords
        v = vec3(c2.coords, c )
        v1 = vec3(c3.coords, c )
        s = [ v[0]+v1[0], v[1]+v1[1], v[2]+v1[2] ]
        mag= b_length / math.sqrt( s[0]*s[0] + s[1]*s[1] + s[2]*s[2])

        coordsH = [c[0] + (s[0]*mag), c[1] + (s[1]*mag), c[2] + (s[2]*mag)]

        atomic_number, type = self.type_added_hydrogen(atom)
        return [(coordsH, atom, atomic_number, type)]


    def add_sp_hydrogen(self, atom, b_length):
        """ """
        b = atom.bonds[0]
        c = atom.coords
        if b.atom1 == atom:
            c2 = b.atom2.coords
        else:
            c2 = b.atom1.coords

        v = vec3(c2, c, b_length)
        coordsH = [ c[0]+v[0], c[1]+v[1], c[2]+v[2] ]
        
        atomic_number, type = self.type_added_hydrogen(atom)
        return [(coordsH, atom, atomic_number, type)]


    def add_vinyl_hydrogens(self, atom, b_length):
        """ """

        # c2 is neigbour of c1
        c2 = atom.bonds[0].atom1
        if c2==atom:
            c2= atom.bonds[0].atom2

        # c3 is neigbour of c2 different from c1
        c3 = c2.bonds[0].atom1
        if c3==c2: c3 = c2.bonds[0].atom2

        if c3==atom:
            c3 = c2.bonds[1].atom1
            if c3 == c2:
                c3 = c2.bonds[1].atom2

        # c4 is neighbor of c2 different from c1 and c3
        c4 = c2.bonds[0].atom1
        if c4==c2:
            c4 = c2.bonds[0].atom2
        if c4 == atom or c4 == c3:
            c4 = c2.bonds[1].atom1
            if c4 == c2:
                c4 = c2.bonds[1].atom2
            if c4==atom or c4 == c3:
                if len(c2.bonds)>2:
                    c4 = c2.bonds[2].atom1
                    if c4==c2:
                        c4 = c2.bonds[2].atom2
                else:
                    c4 = None
        # here we diverge from Babel1.6 because they use H atoms
        # that might have laready been added. We will only add them
        # later. Therefore c2 might not have 3 neighbors
        c = atom.coords
        v = vec3(c3.coords, c2.coords, b_length)
        cH1 = [ c[0]+v[0], c[1]+v[1], c[2]+v[2] ]
        if c4 is not None and len(c2.bonds)==3:
            v1 = vec3(c4.coords, c2.coords, b_length)
            cH2 = [ c[0]+v1[0], c[1]+v1[1], c[2]+v1[2]]
        else:
            # we build second H like a sp2_hydrogen using c2,c,H
            v = vec3(c2.coords, c)
            v1 = vec3(cH1, c, b_length )
            s = [ v[0]+v1[0], v[1]+v1[1], v[2]+v1[2] ]
            mag= b_length / math.sqrt( s[0]*s[0] + s[1]*s[1] + s[2]*s[2])
            cH2 = [c[0] + (s[0]*mag), c[1] + (s[1]*mag), c[2] + (s[2]*mag)]

        atomic_number, htype = self.type_added_hydrogen(atom)

        return [ (cH1, atom, atomic_number, htype),
                 (cH2, atom, atomic_number, htype) ]

    
    def count_missing_hydrogens(self, atoms):
        """ """
        missing = 0
        converter = TypeConverter("HAD")
        
        for a in atoms:
            temp_type = converter.convert(a.babel_type, 'zero')

            if temp_type == 0:
                print "Unable to assign valence to atom %s type = %s" % \
                      (a.full_name(), a.babel_type)

            type_valence = int(temp_type)
            val = len(a.bonds)
            if val < type_valence and val > 0:
                missing = missing + type_valence - val

        return missing


    def count_missing_bo_hydrogens(self, atoms):
        """ """
        missing = 0
        for a in atoms:
            type_valence = 0

            if a.babel_atomic_number==6:
                type_valence = 4

            elif a.babel_atomic_number==7:
                if a.babel_type=='N2' and len(a.bonds)==1:
                    type_valence = 2
                elif a.babel_type=='N3+':
                    type_valence = 4
                else:
                    type_valence = 3

            elif a.babel_atomic_number==8:
                if a.babel_type=="O-" or a.babel_type=="O2":
                    type_valence = 1
                else:
                    type_valence = 2
  
            attached = self.count_attached_bonds(a)
            
            a._redo = 0
            to_add = 0
            if len(a.bonds) < type_valence and len(a.bonds) > 0:
                to_add = type_valence - attached
                if to_add > 0:
                    a._redo = to_add
                    missing = missing + to_add

        return missing


    def count_attached_bonds(self, atom):
        """ """
        bonds = 0.0;
        for b in atom.bonds:
            if b.bondOrder == 5:
                bonds = bonds + 1.5
            elif b.bondOrder == 'aromatic':
                bonds = bonds + 1.5
            elif b.bondOrder == 'amide':
                bonds = bonds + 1.5
            else:
                bonds = bonds + b.bondOrder
        return int(bonds)


if __name__ == '__main__':
    import pdb, sys
    from cycle import RingFinder
    from bo import BondOrder
    from atomTypes import AtomHybridization
    from aromatic import Aromatic
    
    from MolKit.pdbParser import NewPdbParser
    parser = NewPdbParser("/tsri/pdb/struct/%s.pdb"%sys.argv[1])
    mols = parser.parse()
    mol = mols[0]
    mol.buildBondsByDistance()
    allAtoms = mol.chains.residues.atoms
    bonds = allAtoms.bonds[0]

    print "assigning atom types"
    babel = AtomHybridization()
    babel.assignHybridization(allAtoms)

    print "looking for rings"
    rings = RingFinder()
    rings.findRings(allAtoms, bonds)

    print "assigning bond order"
    bo = BondOrder()
    #pdb.run("bo.assignBondOrder(allAtoms, bonds)")
    bo.assignBondOrder(allAtoms, bonds)

    print "looking for aromatic rings"
    arom = Aromatic(rings)
    arom.find_aromatic_atoms(allAtoms)
    
    print "done"
##      db = filter(lambda x:x.bondOrder==2, bonds)
##      for b in db:
##          print b
    
    addh = AddHydrogens()
    #pdb.run("hat = addh.addHydrogens(allAtoms)")
    hat = addh.addHydrogens(allAtoms)
    
    from MolKit.molecule import Atom, Bond
    for a in hat:
        atom = Atom('H', a[1].parent, top=a[1].top)
        atom._coords = [ a[0] ]
        atom.segID = a[1].segID
        atom.hetatm = 0
        atom.alternate = []
        atom.element = 'H'
        atom.number = -1
        atom.occupancy = 1.0
        atom.conformation = 0
        atom.temperatureFactor = 0.0
        atom.babel_atomic_number = a[2]
        atom.babel_type = a[3]
        atom.babel_organic=1
        bond = Bond( a[1], atom )

    from Pmv.moleculeViewer import MoleculeViewer
    mv = MoleculeViewer()
    mv.addMolecule(mol)
    mv.lines( mol )



import sys, string
from mglutil.regression import testplus

class Atom:

    def __init__(self, coords):
        self.coords = coords
        self.bonds = []

class Bond:

    def __init__(self, atom1, atom2):
        self.atom1 = atom1
        self.atom2 = atom2
        self.atom1.bonds.append(self)
        self.atom2.bonds.append(self)


def getGraphFromFile(filename):
    f = open(filename)
    nba = int(f.readline())
    atoms = []
    for i in range(nba):
        data = string.split(f.readline())
        x,y,z = map( float, data)
        atoms.append(Atom((x,y,z)))
    bonds = []
    nbb = int(f.readline())
    for i in range(nbb):
        data = string.split(f.readline())
        at1,at2 = map( int, data)
        bonds.append( Bond( atoms[at1], atoms[at2] ) )
    return atoms, bonds


def test_findCycles4():
    # test that atoms in cycles are ordered on taxol 
    atoms, bonds = getGraphFromFile('txp.graph')
    from PyBabel.cycle import RingFinder
    rings = RingFinder()
    rings.findRings2(atoms, bonds)

    def hasBond(a1, a2, cycleHasBond):
        for b in cyclebonds:
            if b.atom1==a1 and b.atom2==a2:
                return 1
            elif b.atom1==a2 and b.atom2==a1:
                return 1
        return 0
    
    def isOrdered(cycleatoms, cyclebonds):

        def hasBond(a1, a2, cyclebonds):
            for b in cyclebonds:
                if b.atom1==a1 and b.atom2==a2:
                    return 1
                elif b.atom1==a2 and b.atom2==a1:
                    return 1
            return 0

        start = cycleatoms[0]
        for end in cycleatoms[1:]:
            if not hasBond(start, end, cyclebonds):
                return 0
            start = end
        return 1
    
    for r in rings.rings:
        assert isOrdered(r['atoms'], r['bonds'])


def test_findCycles1():
    # test cycle detection on taxol 
    atoms, bonds = getGraphFromFile('txp.graph')
    from PyBabel.cycle import RingFinder
    rings = RingFinder()
    rings.findRings2(atoms, bonds)
    assert rings.ringCount==6
    
def test_findCycles2():
    # test cycle detection on modified taxol to have 5 membered ring
    atoms, bonds = getGraphFromFile('fuzedRings2.graph')
    from PyBabel.cycle import RingFinder
    rings = RingFinder()
    rings.findRings2(atoms, bonds)
    assert rings.ringCount==4

    
def test_findCycles3():
    # test cycle detection on taxol merged rings only
    atoms, bonds = getGraphFromFile('fuzedRings1.graph')
    from PyBabel.cycle import RingFinder
    rings = RingFinder()
    rings.findRings2(atoms, bonds)
    assert rings.ringCount==4


harness = testplus.TestHarness( "PyBabel_cycles",
                                funs = testplus.testcollect( globals()),
                                )

if __name__ == '__main__':
    testplus.chdir()
    print harness
    sys.exit( len( harness))



############################################################################
#
# Author: Michel F. SANNER
# Reimplemented from Babel v1.6 from Pat Walters and Math Stahl
#
# Copyright: M. Sanner TSRI 2000
#
#############################################################################
#
# $Id: cycle.py,v 1.8 2003/09/04 23:53:56 lindy Exp $
#

"""
This file implements the RingFinder class that can be used to identify
rings in molecules. When rings are nested the smalest rings are reported.
The algorithms in here are different from thoses in Babel1.6. They might
fail to report all rings for rings in which any atom of a ring belongs
also to more than 1 ring.
findRings2 is more robust and should reports all the smallest cycles. It is
more expansive too. Should be used for small molecules

example:
    
    >>> r = RingFinder()
    >>> r.findRings2(atoms, bonds)
    >>> r.printRings()

      atoms has to be a list of Atom objects
      Atom:
          a.coords : 3-sequence of floats
          a.bonds : list of Bond objects
      Bond:
          b.atom1 : instance of Atom
          b.atom2 : instance of Atom

      after completion the RingFinder object has the following members:
          ringCount: number of rings
          rings : a list of rings. A ring is a dictionary with 2 keys
                  'atoms' and 'bonds'. The atoms in rings['atoms'] are
                  ordered along the cycle.
          allRingAtoms: list of atoms that are in rings(atoms may appear twice)
          allRingBonds: list of bonds that are in rings(bonds may appear twice)

      In addition:
          atoms involved in rings have a member 'rings' that is a list of
          rings they belong to (0-based list of integers)

Michel Sanner April 2000
"""




class RingFinder:
    """ """
    def __init__(self):
        """ """
        self.num = 0
        self.rings = []


    def tagOneAtomBond(self, atom, inbond):
        # tag all bond that cannot be in a ring by adding them as keys to
        # self.bondInCycle[bond], value is set to 2 (no particular meaning)

        # since findRings2 will oly consider atoms for which at least one
        # bond is not yet in a cycle, pretending that the only bond to an
        # atom is already in a cycle will pevent it from being considered
        if len(atom.bonds)==1:
            bond = atom.bonds[0]
            self.bondInCycle[bond] = 2
            #print 'REMOVE ',atom.name,bond.atom1.name,'-',bond.atom2.name
            if bond!=inbond:
                atom2 = bond.atom1
                if atom2==atom:
                    atom2=bond.atom2
                self.tagOneAtomBond(atom2, bond)
        else:
            num = 0
            for b in atom.bonds:
                if not self.bondInCycle.has_key(b):
                    num = num + 1
                    bond = b
            if num == 1:
                #print 'REMOVE ',atom.name,bond.atom1.name,'-',bond.atom2.name
                self.bondInCycle[bond] = 2
                atom2 = bond.atom1
                if atom2==atom:
                    atom2=bond.atom2
                self.tagOneAtomBond(atom2, bond)

                
    def findRings2(self, atoms, bonds, maxSize=20):
        # EXPANSIVE but robust
        # for each atom for which at least one bond is not yet in a cycle
        # find the smallest cycle. Then check if the cycle is already known
        # if maxSize is specified, only ring of size maxSize will be found
        
        self.bondInCycle = {} # key is bond, is created when bond has been
                              # traversed (persistant across search over atoms
        self._rings = []      # dict. with atoms as keys per cycle
        self.allRingAtoms = {} # dict with keys being all atoms in cycles
        self.allRingBonds = {} # dict with keys being all bonds in cycles
        self.ringCount = 0     # number of cycles
        maxLevel = maxSize/2 + 1

        # tag all bonds that cannot be in cycles, i.e. they connect an atom
        # having only 1 neighbor (recursively)
        #for a in atoms:
        #    self.tagOneAtomBond(a, None)
            
        # loop over all atoms
        for a in atoms:
            # being in a cycle requires at least 2 neighbors
            if len(a.bonds)==1:
                continue
            # count number of bond already in cycles or connecting
            # to an leaf atom (i.e. atom with only 1 neighbor)
            num = 0
            for b in a.bonds:
                atom2 = b.atom1
                if atom2==a: atom2=b.atom2
                if len(atom2.bonds)==1 or self.bondInCycle.has_key(b):
                    num = num + 1

            # if at least on bond not in a cycle find smallest cycle for a
            if num < len(a.bonds):
                #print 'Find smallest Ring for', a.name, num, len(a.bonds)
                # ra is a list of atoms in cycle
                # rb is a list of bonds in cycle
                #print 'Find smallest cycle for', a.name
                ra, rb = self.findSmallestRing(a, maxLevel)
                #print 'Found cycle:', ra
                # check is that cycle has been found before
                if len(ra):
                    same = 0
                    for r in self._rings:
                        if len(ra)==len(r):
                            same = 1
                            for a in ra:
                                if not r.has_key(a):
                                    same = 0
                                    break
                            if same==1:
                                break

                    # if it hasn't be found before, add it
                    if not same:
                        # atoms in ra have to be sorted along the cycle
                        ras = [ra[0]]
                        # Take all odds first going trough list toward end
                        for i in range(1, len(ra), 2):
                            ras.append(ra[i])
                        # take all even going through list backwards and
                        # starting a last of cycles woth even number of atoms
                        # and one before last of rinfs with odd number of at.
                        l = len(ra)
                        if (l/2*2) != l:
                            end = l-1 # odd -> skip last
                        else:
                            end = l-2 # end, use last
                        for i in range(end, 1, -2):
                            ras.append(ra[i])
                        ra = ras
                        
                        # build the dict of atoms for that cycle
                        d = {}
                        ringnum = len(self._rings)
                        for a in ra:
                            d[a] = 1
                            if not hasattr(a, 'rings'):
                                a.rings = [ringnum]
                            else:
                                a.rings.append(ringnum)
                        # add it to _rings
                        self._rings.append(d)
                        # add the cycle to self.rings
                        self.rings.append( {'atoms':ra, 'bonds':rb } )
                        # update dict of all atoms in cycles
                        self.allRingAtoms.update(d)
                        # update dict of all bonds in cycles
                        for b in rb:
                            self.allRingBonds[b] = 1
                        # increment cycl counter
                        self.ringCount = self.ringCount + 1

##                          print 'RING ======================'
##                          for a in ra:
##                              print a.name
##                          for b in rb:
##                              print b.atom1.name,'-',b.atom2.name
##                          print 'END RING ======================'
        self.allRingAtoms = self.allRingAtoms.keys()
        self.allRingBonds = self.allRingBonds.keys()


    def findSmallestRing(self, root, maxLevel):
        # each new generation adds a level in the width first traversal
        level = 0
        # each level has a stack of atoms
        stacks = []
        # each level has a dict of atom:bond telling through which bond we
        # came to this atom
        bndDicts = []
        # dict of atoms used to check if an atom has been seen before
        atinstack = {}
        # dict of bonds used to check if a bond has been seen before
        bondseen = {}

        # bond through which we came (None for root)
        bond = None
        atinstack[root] = 1
        # list of atoms a this level
        levelstack = []
        stacks.append(levelstack)
        # dict of atom:bond for this level
        levelDict = {}
        bndDicts.append(levelDict)

        # add children of root to this level's stack and dict
        for b in root.bonds:
            bondseen[bond] = 1
            atom2 = b.atom1
            if atom2==root:
                atom2 = b.atom2
            if len(atom2.bonds)>1:
                #print 'adding to stack 0', atom2.name
                levelDict[atom2] = b
                levelstack.append(atom2)
                atinstack[atom2] = 1

        maxLen = len(levelstack) # number of bonds with level 0
        stackPtr = 0

        # width first traversal, i.e. loop over stack adding levels
        while stackPtr < maxLen:
            # add a level
            levelstack = []
            stacks.append(levelstack)
            levelDict = {}
            bndDicts.append(levelDict)
            #print "Looping over LEVEL: ", level
            # loop over atoms at this level
            for levelroot in stacks[level]:
##                  if len(levelroot.bonds)==1:
##                      continue
                # find bond through with we came to levelroot atom
                bond = bndDicts[level][levelroot]
                # if already seen get next level root

##                  if bondseen.has_key(bond):
##                      continue

##                  if len(bond.atom1.bonds)==1 or len(bond.atom2.bonds)==1:
##                      continue # this bond cannot be in a cycle
##                  if self.bondInCycle.has_key(bond) and \
##                     self.bondInCycle[bond]==2:
##                      continue
                # else make this bond as seen
                bondseen[bond] = 1
                #print 'level ROOT', levelroot.name, bond, atinstack.has_key(levelroot)
                # add children of levelroot to the current level
                for b in levelroot.bonds:
                    # except for the parent of levelroot
                    if b is bond:
                        continue
                    atom2 = b.atom1
                    if atom2==levelroot:
                        atom2 = b.atom2
                    if len(atom2.bonds)==1:
                        continue
                    
                    #print levelroot.name, atom2.name
                    # if the child of levelroot is instack we found a cycle
                    # so we start back tracking from both sides through the
                    # levels until we reach a common atom. If that atom is
                    # root we have the smallest cycle for root, else we
                    # continue

                    # 2 cases are possible: either we have a even or an odd
                    # number of atoms in the cycle. For even numbers atom2
                    # is found in atinstack which is associated with this
                    # level. else atom2 should be found in
                    # bndDicts[level]

                    if atinstack.has_key(atom2):
                        #print 'CYCLE', atom2.name
                        # even number of atoms
                        #print 'instack'
                        #for a in atinstack:
                        #    print a.name
                        if levelDict.has_key(atom2):
                            #print 'EVEN ******************'
                            # cycle with even number of atoms
                            # b1 is the bond from which we arrived at atom2
                            # previousely at this level
                            b1 = levelDict[atom2]
                            at1 = levelroot
                            at2 = b1.atom1
                            if at2==atom2: at2 = b1.atom2

                            #print 'ringAtoms1', atom2.name, at1.name, at2.name
                            ringAtoms = [atom2, at1, at2]
                            ringBonds = [b, b1]
                            # backtrack through level
                        elif bndDicts[level].has_key(atom2):
                            # odd number of atoms in cycle
                            # b1 is the bond from which we arrived at atom2
                            # previousely at this level
                            #print 'ODD ******************'
                            at1 = levelroot
                            at2 = atom2
                            #print 'ringAtoms2', at1.name, at2.name

                            ringAtoms = [at1, at2]
                            ringBonds = [b]
                        else:
                            continue

                        # backtrack
                        for i in range(level,-1,-1):
                            #print 'level:', i
                            b1 = bndDicts[i][at1]
                            other1 = b1.atom1
                            if other1==at1: other1 = b1.atom2
                            #print other1.name
                            b2 = bndDicts[i][at2]
                            other2 = b2.atom1
                            if other2==at2: other2 = b2.atom2
                            #print other2.name
                            at1 = other1
                            at2 = other2
                            ringBonds.append(b1)
                            ringBonds.append(b2)
                            ringAtoms.append(other1)
                            if other1!=other2:
                                ringAtoms.append(other2)
                            else:
                                break
                        if other1==root or other2==root:
                            for b in ringBonds:
                                self.bondInCycle[b] = 1
                            return (ringAtoms, ringBonds)

                    else:
                        #print 'adding to stack', atom2.name, len(levelstack)
                        levelstack.append(atom2)
                        levelDict[atom2] = b
                        atinstack[atom2] = 1

            level = level + 1
            if level==maxLevel:
                return [],[]
            maxLen = maxLen + len(levelstack)
            stackPtr = stackPtr + 1
        return [],[]
    

    def backtrack(self, atom1, atom2, b):
        """go up ancestor tree until first common parent is found"""
        ringAtoms = [ atom2 ]
        ringbonds = [ b ]
        b._ring_seen = 1
        while atom1!=atom2:
            ringAtoms.append( atom1 )
            ringbonds.append( atom1._ring_ancestor_bond )
            atom1 = atom1._ring_ancestor
        return { 'atoms':ringAtoms, 'bonds':ringbonds }


    def tag_neighbors(self, atom1, bond):
        """ """
        atom2 = bond.atom1
        if atom2==atom1: atom2 = bond.atom2
        #print 'atom2', atom2.name, bond
        if hasattr(atom2,'_ring_ancestor'):
            self._rings.append( self.backtrack(atom1, atom2, bond) )
            #print '\n********************* ring found'
            #for a in self._rings[-1]['atoms']:
                #print a.name
            return

        atom2._ring_ancestor = atom1
        atom2._ring_ancestor_bond = bond
        bond._ring_seen = 1
        for b in atom2.bonds:
            #FIXME: we are supposed to skip di-sulfite bridges here
            if hasattr(b, '_ring_seen') and b._ring_seen:
                continue
            self.tag_neighbors( atom2, b )
        #print 'step back from', atom2.name, bond
        
                
    def findRings(self, atoms, bonds):
        """method to find cycles in molecules, first we simply tag all atoms
        and bonds in rings using a depth first traversal then we call
        checkRings for identifying the smallest cycles when fused rings are
        present
        """

        i = 0
        for b in bonds:
            b._ring_seen = 0
            
        self._rings = []

        # first try with leaf atoms (i.e only 1 neighbor)
        done = 0
        for a in atoms:
            if len(a.bonds)==1:
                a._ring_ancestor = None
                a._ring_ancestor_bond = None
                #print 'start at ',a, a.bonds[0]
                self.tag_neighbors(a, a.bonds[0])
                done = 1
                break

        #print 'AAAA', done
        
        # for molecules with all atoms in cycles we did not find a leaf
        if not done:
            for a in atoms:
                if not hasattr(a, '_ring_ancestor'):
                    l = len(a.bonds)
                    if l==2:
                        atom_before = a.bonds[0].atom1
                        if atom_before==a: atom_before=a.bonds[0].atom2
                        a._ring_ancestor = atom_before
                        a._ring_ancestor_bond = a.bonds[0]
                        self.tag_neighbors(a, a.bonds[1])

                        if a.bonds[0]._ring_seen:
                            continue
                        atom_before = a.bonds[1].atom1
                        if atom_before==a: atom_before=a.bonds[1].atom2
                        a._ring_ancestor = atom_before
                        a._ring_ancestor_bond = a.bonds[0]
                
##          for a in atoms:
##              if not hasattr(a, '_ring_ancestor'):
##                  l = len(a.bonds)
##                  if l==0 or l > 2: continue
##                  if l==2:
##                      atom_before = a.bonds[0].atom1
##                      if atom_before==a: atom_before=a.bonds[0].atom2
##                      a._ring_ancestor = atom_before
##                      a._ring_ancestor_bond = a.bonds[0]
##                      self.tag_neighbors(a, a.bonds[1])

##                      if a.bonds[0]._ring_seen:
##                          continue
##                      atom_before = a.bonds[1].atom1
##                      if atom_before==a: atom_before=a.bonds[1].atom2
##                      a._ring_ancestor = atom_before
##                      a._ring_ancestor_bond = a.bonds[0]
##                  else:
##                      a._ring_ancestor = None
##                      a._ring_ancestor_bond = None
##                      self.tag_neighbors(a, a.bonds[0])

        for a in atoms:
            if hasattr(a,'_ring_ancestor'):
                delattr(a, '_ring_ancestor')
            if hasattr(a,'_ring_ancestor_bond'):
                delattr(a, '_ring_ancestor_bond')
        for b in bonds:
            delattr(b, '_ring_seen')

        self.checkRings()
        delattr(self, '_rings')

        
    def smallestCycle(self, atom):
        """
        find smalest cycle containing starting at atom and traversing
        tree in breadth first order. When an atom with _ancestor is found
        we backtrack and both sides to build listst of atoms and bonds.
        It is possible that the first cycle does not contain the initial
        atom.
        """
        if self._result:
            return self._result
        
        for b in atom.bonds:
            if self._result:
                return self._result
            if not hasattr(b, '_ring'): continue
            if b._seen: continue
            atom2 = b.atom1
            if atom2==atom: atom2 = b.atom2
            if atom2._ancestor:
                l1 = [atom2]
                atom2.rings.append( self.ringCount )
                l2 = [atom]
                atom.rings.append( self.ringCount )
                b1 = [b]
                b2 = []
                a2 = atom2
                a1 = atom
                done = 0
                while not done:
                    b1.append(a2._ancestor_bond)
                    a = a2._ancestor
                    if a != l2[-1]:
                        l1.append(a)
                        a.rings.append( self.ringCount )
                        a2 = a
                    else:
                        break

                    b2.append(a1._ancestor_bond)
                    a = a1._ancestor
                    if a != l1[-1]:
                        l2.append(a)
                        a.rings.append( self.ringCount )
                        a1 = a
                    else:
                        break
                    
                l2.reverse()
                b2.reverse()
                self._result = (l2+l1, b2+b1)
                return
            else:
                atom2._ancestor = atom
                atom2._ancestor_bond = b
                self.stack.append(atom2)
                b._seen = 1
           
        if len(self.stack):
            a = self.stack[0]
            self.stack.remove(a)
            self.smallestCycle(a)

        return self._result


    def checkRings(self):
        """
        this functions uses the rings found by findRings to identify smalest
        cycles in structure.

        After this method was called this object as the following new members:
          rings: a list of dictionnaries {'atoms':list of atoms, 'bonds': list}
          allRingAtoms: list of all atoms in rings
          allRingBonds: list of all bonds in rings
          ringCount: number of rings
        """
        self.rings = []
        self.allRingAtoms = []
        self.allRingBonds = []
        self.ringCount = 0
        for ring in self._rings:
            self.allRingAtoms = self.allRingAtoms + ring['atoms']
            self.allRingBonds = self.allRingBonds + ring['bonds']

        for a in self.allRingAtoms: a.rings = []
        for b in self.allRingBonds: b._ring = 1

        # breadth first walking over atoms and bonds in rings
        # tag atoms in shortest cycle
        for a in self.allRingAtoms:
            if len(a.rings)>0: continue

            self.stack = []
            for b in self.allRingBonds: b._seen = 0
            for at in self.allRingAtoms: at._ancestor = None
            a._ancestor_bond = None
            self._result = None

            ratoms, rbonds = self.smallestCycle(a)

            if a in ratoms:
                self.ringCount = self.ringCount+1
                self.rings.append( {'atoms':ratoms, 'bonds':rbonds} )
            else:
                for at in ratoms:
                    at.rings = at.rings[:-1]

        # clean up
        for a in self.allRingAtoms:
            if hasattr(a, '_ancestor'): delattr(a, '_ancestor')
            if hasattr(a, '_ancestor_bond'): delattr(a, '_ancestor_bond')

        for b in self.allRingBonds:
            if hasattr(b, '_ring'): delattr(b, '_ring')
            if hasattr(b, '_seen'): delattr(b, '_seen')

        if hasattr(self, '_result'): delattr(self, '_result')

        
    def printRings(self):
        """ """
        if not hasattr(self, 'rings'):
            return
        i = 0
        for r in self.rings:
            print 'RING ',i
            for j in range(len(r['atoms'])):
                a = r['atoms'][j]
                b = r['bonds'][j]
                print '%10s %4d %s'%(a.name, a.number, repr(b))
            i = i + 1

            
if __name__ == '__main__':
    import pdb, sys
    from MolKit.pdbParser import NewPdbParser
    parser = NewPdbParser("/tsri/pdb/struct/%s.pdb"%sys.argv[1])
    mols = parser.parse()
    mol = mols[0]
    mol.buildBondsByDistance()
    allAtoms = mol.chains.residues.atoms

    print 'Looking for rings'
    r = RingFinder()
    bonds = (allAtoms.bonds)[0]
    r.findRings(allAtoms, bonds)
    r.printRings()


    from MolKit.pdbParser import NewPdbqParser
    parser = NewPdbqParser("./txp.pdbq")
    mols = parser.parse()
    mol = mols[0]
    mol.buildBondsByDistance()
    allAtoms = mol.chains.residues.atoms


    print 'Looking for rings ...'
    r = RingFinder()
    print "Done"
    bonds = (allAtoms.bonds)[0]
    r.findRings(allAtoms, bonds)

    r.printRings()
